"""Keep reference to running asyncio tasks"""

import asyncio
from collections.abc import Coroutine

from aiobp import log

__tasks: set[asyncio.Task] = set()  # to avoid garbage collection by holding reference


async def exception_handler(coroutine: Coroutine, name: str) -> Coroutine:
    """Log unhandled exception in asyncio task

    This is needed to get exceptions immediately (otherwise they are displayed on shutdown).
    """
    try:
        await coroutine
    except Exception as error:  # noqa: BLE001 - yes, we want to catch it and log it
        log.critical('Unhandled exception in task "%s"', name, exc_info=error)


def create_task(coroutine: Coroutine, name: str) -> asyncio.Task:
    """Create task and keep reference until the task is done

    Argument "name" is optional in asyncio.task(), however we require it
    to make our code easier to debug.
    """
    task = asyncio.create_task(exception_handler(coroutine, name=name), name=name)
    __tasks.add(task)
    task.add_done_callback(__tasks.discard)
    return task



"""Add dev and trace logging shortcuts"""

import functools
import logging

dev = functools.partial(logging.log, 1)
debug = logging.debug
info = logging.info
warning = logging.warning
error = logging.error
critical = logging.critical
exception = logging.exception
trace = logging.exception



"""Customize Python logger"""

import logging
import logging.handlers
import sys
from types import TracebackType
from typing import ClassVar, Optional


class LoggingConfig:
    """Loging configuration"""

    level: str = "DEBUG"
    filename: Optional[str] = None


class Color:
    """Console colors"""

    black = "\x1b[0;30m"
    brown = "\x1b[0;31m"
    dark_green = "\x1b[0;32m"
    orange = "\x1b[0;33m"
    dark_blue = "\x1b[0;34m"
    purple = "\x1b[0;35m"
    teal = "\x1b[0;36m"
    grey = "\x1b[0;37m"
    red = "\x1b[0;91m"
    green = "\x1b[0;92m"
    yellow = "\x1b[0;93m"
    blue = "\x1b[0;94m"
    magenta = "\x1b[0;95m"
    cyan = "\x1b[0;96m"
    white = "\x1b[0;97m"
    reset = "\x1b[0m"


class PrefixExceptionFormatter(logging.Formatter):
    """Add date in front of each line of multiline exception/stack trace"""

    def _prefix_text(self, prefix: str, stack: str, color: Color = Color.brown) -> str:
        """Prefix lines with date so we can easily grep by date and time"""
        return "\n".join([f"{prefix} {color}{line}{Color.reset}" for line in stack.split("\n")][1:])


class ColorFormatter(PrefixExceptionFormatter):
    """Color formatter for console logging"""

    severity_color: ClassVar[dict[int, str]] = ({
        logging.DEBUG: Color.grey,
        logging.INFO: Color.white,
        logging.WARNING: Color.orange,
        logging.ERROR: Color.red,
        logging.CRITICAL: Color.magenta,
    })

    def format(self, record: logging.LogRecord) -> str:
        """Format message according to its severity level"""
        color = ColorFormatter.severity_color.get(record.levelno, Color.blue)
        record.message = f"{color}{record.getMessage()}{Color.reset}"
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        s = self.formatMessage(record)
        if record.exc_info and not record.exc_text:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self._prefix_text(record.asctime, record.exc_text)
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self._prefix_text(record.asctime, self.formatStack(record.stack_info))
        return s


def add_devel_log_level() -> None:
    """Register new logging level for development purposes"""
    logging.addLevelName(1, "DEVEL")


def log_unhandled_exception(type_: type[BaseException], value: BaseException, traceback: TracebackType) -> None:
    """Log unhandled exceptions"""
    if issubclass(type_, KeyboardInterrupt):
        sys.__excepthook__(type_, value, traceback)
        return

    logging.critical("Uncaught exception", exc_info=(type_, value, traceback))


def setup_logging(config: Optional[LoggingConfig] = None) -> None:
    """Setups Python logger"""
    if config is None:
        config = LoggingConfig()

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(
        ColorFormatter(f"%(asctime)s [{Color.green}%(filename)s:%(lineno)d{Color.reset}] %(message)s"),
    )
    handlers = [console_handler]

    if config and config.filename is not None:
        file_handler = logging.handlers.WatchedFileHandler(filename=config.filename)
        file_handler.setFormatter(
            PrefixExceptionFormatter("%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"),
        )
        handlers.append(file_handler)

    logging.basicConfig(level=config.level if config else None, handlers=handlers)

    sys.excepthook = log_unhandled_exception



"""Load configuration from YAML file"""

from pathlib import Path
from typing import Annotated, Optional, get_origin

import yaml

from .annotations import ConfigOption, get_options, parse_list, set_options
from .exceptions import InvalidConfigFile, InvalidConfigImplementation


def loader(config_class: type[Annotated], filename: Optional[str] = None) -> Annotated:
    """Load configuration from YAML file"""
    config = get_options(config_class)

    if filename is None:
        return set_options(config_class(), config)

    file = Path(filename)
    with file.open(encoding="utf-8") as fp:
        conf = yaml.safe_load(fp)

    if not isinstance(conf, dict):
        error = "YAML configuration structure must be dict[str, dict[str, any]]"
        raise InvalidConfigFile(error)

    for section_name, options in config.items():
        if not isinstance(options, dict):
            error = f'Class "{config_class.__name__}" can\'t have direct option "{section_name}"'
            raise InvalidConfigImplementation(error)

        conf_section = conf.get(section_name, {})

        for option_name, option in options.items():
            if not isinstance(option, ConfigOption):
                error = f'"{section_name}" can have only scalar attributes, not subsection "{option_name}"'
                raise InvalidConfigImplementation(error)

            value = conf_section.get(option_name, option.value)
            if isinstance(value, Exception):
                options[option_name] = value
            else:
                typed_list = get_origin(option.type) is list
                options[option_name] = parse_list(option.type, value) if typed_list else option.type(value)

    return set_options(config_class(), config)



"""Configuration structure from class annotations

Given config:

class ServiceConfig:
    host: str
    port: int = 8080


class LogConfig:
    level: str = "DEBUG"
    filename: str


class Config:
    service: ServiceConfig
    log: LogConfig


Method get_options inspects classes recursively and return configuration map as
nested dicts of options/attributes (keys) having type and default value (vals):

config_map = get_options(Config)
{
    # module name
    "service": {
        # option name: type, default value or AttributeError if default is not provided
        "host": (str, AttributeError),
        "port": (int, 8080),
    },
    "log": {
        "level": (str, "DEBUG"),
        "filename": (str, AttributeError),
    },
}


Loader class constructor will use configuration map, fill values from configuration
file (override default values) and use set_options method to make instances of nested
configuration classes:


class Loader:
    def __init__(self, filename):
        config_map = get_options(Config)

        # parse filename and fill values

        set_options(self, config_map)


Finally what we need to use in our microservice:

config = Config(Loader("some.file"))
"""

import contextlib
from typing import Annotated, Any, NamedTuple, Optional, Union, get_args, get_type_hints

from .exceptions import InvalidConfigFile, InvalidConfigImplementation


class ConfigOption(NamedTuple):
    type: Any
    value: Any

ConfigMap = dict[str, Union["ConfigMap", ConfigOption]]


CONFIG_TYPES = (
    bool,
    float,
    int,
    str,
    Optional[bool],
    Optional[float],
    Optional[int],
    Optional[str],
    list[bool],
    list[float],
    list[int],
    list[str],
)


def get_options(config_class: type[Annotated]) -> ConfigMap:
    """Extract configuration map recursively"""
    config: ConfigMap = {}

    options = get_type_hints(config_class)
    if not options:
        error = f'Configuration class "{config_class}" has no configuration options'
        raise InvalidConfigImplementation(error)

    for option_name, option_type in options.items():
        if option_type not in CONFIG_TYPES:
            config[option_name] = get_options(option_type)
            continue

        try:
            default_value = getattr(config_class, option_name)
        except AttributeError as error:
            default_value = error
        config[option_name] = ConfigOption(option_type, default_value)

    return config


def parse_list(value_type: type, value: Union[str, list[Any]]) -> list[Any]:
    """Validate list[Any] configuration option"""
    if isinstance(value, str):
        value = [raw.strip() for raw in value.split(",")]

    for option_type in get_args(value_type):
        with contextlib.suppress(ValueError, TypeError):
            return [option_type(raw) for raw in value]

    error = f"invalid configuration value for {value_type}: {value}"
    raise InvalidConfigFile(error)


def set_options(
    config: Annotated,
    values: ConfigMap,
    path: Optional[list[str]] = None,
) -> Annotated:
    """Make instances of configuration classes and set values"""
    for option_name, option_type in get_type_hints(config).items():
        if option_type not in CONFIG_TYPES:
            path = [option_name] if path is None else [*path, option_name]
            value = set_options(option_type(), values.get(option_name), path)
        else:
            value = values.get(option_name)
            if isinstance(value, ConfigOption):  # option was not set by configuration file loader
                value = value.value
            if isinstance(value, Exception):
                error = f'missing option "{option_name}" in section "{".".join(path)}"'
                raise InvalidConfigFile(error) from value

        setattr(config, option_name, value)

    return config



"""Configration exceptions"""

class InvalidConfigFile(BaseException):
    """Invalid configuration file"""


class InvalidConfigImplementation(BaseException):
    """Invalid usage of annotations for configuration"""
    # you shouldn't catch this exception
    # it indicates invalid code, not invalid configuration file itself



"""Configuration provider"""

import sys
from pathlib import Path

from .exceptions import InvalidConfigFile


def sys_argv_or_filenames(*filenames: str) -> str:
    """Return usable configuration filename"""
    if len(sys.argv) > 1:
        filename = sys.argv[1]
        if not Path(filename).is_file():
            error = f'Provided filename "{filename}" not found'
            raise InvalidConfigFile(error)

        return filename

    for filename in filenames:
        if not Path(filename).is_file():
            continue

        return filename

    error = f'None of default filenames found: {", ".join(filenames)}'
    raise InvalidConfigFile(error)


__all__ = ["InvalidConfigFile", "sys_argv_or_filenames"]



"""Run asyncio coroutine as endless service and handle graceful shutdown"""

import asyncio
import contextlib
import signal
import time
from collections.abc import Coroutine
from pathlib import Path
from typing import Any, Callable, Optional, Union

from .logging import log
from .task import create_task

__on_shutdown: list[tuple[Callable[..., Coroutine], list[Any], bool]] = []  # to gracefully close connections


def on_shutdown(
    # instead of ..., there should be generic TypeVar reflected to args
    # however python type system doesn't seem to be capable of that yet
    coroutine: Callable[..., Coroutine],
    args: Optional[list[Any]] = None,
    *,
    after_tasks_cancel: bool = False,
) -> None:
    """Register coroutine to be called on graceful shutdown

    Coroutines are called in LIFO order. By default, coroutines are called
    BEFORE all tasks are canceled and awaited for graceful completion.

    If after_tasks_cancel is set to True then the coroutine is called
    AFTER canceling all tasks and waiting for their possible timeout.
    Useful for ClientSession.close() for example.
    """
    if args is None:
        args = []
    __on_shutdown.append((coroutine, args, after_tasks_cancel))
    log_args = ",".join(repr(arg) for arg in args)
    log_when = "after" if after_tasks_cancel else "before"
    log.debug("Will call %s(%s) on shutdown %s tasks cancelation", coroutine, log_args, log_when)


def log_awaitable(awaitable: Union[asyncio.Task, Coroutine]) -> str:
    """Make good task description for log"""
    coroutine = awaitable.get_coro() if isinstance(awaitable, asyncio.Task) else awaitable
    if not asyncio.iscoroutine(coroutine):
        return repr(coroutine)

    code = coroutine.cr_code

    cwd = Path.cwd().as_posix()
    filename = code.co_filename
    if filename.startswith(cwd):
        filename = filename[len(cwd) + 1 :]

    if awaitable == coroutine:  # it was coroutine, not a task
        return f'<{filename}:{code.co_firstlineno} method "{code.co_name}">'

    return f'<{awaitable.get_name()} {filename}:{code.co_firstlineno} metod "{code.co_name}">'


# Needed up to Python 3.10, when we upgrade to Python 3.11 we can use builtin asyncio.run()
def runner(service: Coroutine, shutdown_timeout: float = 5.0, *, endless: bool = True) -> None:
    """Run given service in asyncio.Task and handle SIGTERM/KeyboardInterrupt

    If endless is set to False then runner shutdown immediately after srvice
    coroutine finishes; otherwise the service is kept running.
    """
    loop = asyncio.get_event_loop()
    # main does:
    # 1. start given service coroutine as task
    # 2. add SIGTERM and SIGINT handlers
    # 3. waits for kill or KeyboardInterrupt
    try:
        loop.run_until_complete(__main(service, endless=endless))
    except Exception:  # noqa: BLE001 - yes, we want to catch it and log it
        log.critical("Unhandled exception in service")
    # graceful_shutdown does:
    # 1. calls one by one all registered coroutines via on_shutdown(...) in LIFO order
    # 2. cancel all tasks in parallel (via gather)
    # 3. await tasks to finish in specified shutdown_timeout
    # 4. calls all registered coroutines via on_shutdown(..., after_tasks_cancel=True) in LIFO order
    loop.run_until_complete(__graceful_shutdown(shutdown_timeout))


async def __shutdown() -> None:
    """Wait for CTRL+C or SIGTERM"""
    shutdown_event = asyncio.Event()
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGTERM, shutdown_event.set)
    loop.add_signal_handler(signal.SIGINT, shutdown_event.set)
    with contextlib.suppress(asyncio.CancelledError):
        await shutdown_event.wait()


async def __main(service: Coroutine, *, endless: bool) -> None:
    """Run main service in task"""
    main_task = create_task(service, "MainTask")

    if endless:  # endless wait for interrupt
        await __shutdown()
    else:  # wait for service coroutine to finish or interrupt
        await asyncio.wait(
            [main_task, create_task(__shutdown(), name="Shutdown")],
            return_when=asyncio.FIRST_COMPLETED,
        )


async def __graceful_shutdown(timeout: float = 5.0) -> None:
    """Gracefully shutdown coroutines and tasks"""
    log.info(f"Graceful shutdown up to {timeout:.3f} s...")
    remains = timeout
    remains = await __shutdown_coroutines(timeout=remains, after_tasks_cancel=False)
    remains = await __shutdown_tasks(timeout=remains)
    remains = await __shutdown_coroutines(timeout=remains, after_tasks_cancel=True)
    if remains > 0:
        log.info("Shutdown completed in %.3f s", (timeout - remains))
    else:
        log.warning("Shutdown not completed within timeout!")


async def __shutdown_coroutines(timeout: float, *, after_tasks_cancel: bool) -> float:
    """Await for on_shutdown callbacks finish with timeout"""
    log.debug(f"Shutting down coroutines {'after' if after_tasks_cancel else 'before'} tasks cancel...")

    for coro, args, when in __on_shutdown[::-1]:
        if when != after_tasks_cancel:
            continue

        start = time.time()
        log_args = ",".join(repr(arg) for arg in args)
        try:
            coroutine = coro(*args)
        except Exception:  # noqa: BLE001 - yes, we want to catch it and log it
            log.trace("Exception in %s(%s)", coro.__name__, log_args)
            took = time.time() - start
            timeout -= took
            continue

        if not asyncio.iscoroutine(coroutine):  # we may get just plain synchronous method
            took = time.time() - start
            timeout -= took
            continue

        try:
            result = await asyncio.wait_for(coroutine, timeout=timeout)
            log.debug("Coroutine finished: %s(%s) -> %r", log_awaitable(coro), log_args, result)
        except asyncio.TimeoutError:
            log.error("Coroutine did not finish in timeout: %s(%s)", log_awaitable(coro), log_args)
        except Exception as error:  # noqa: BLE001 - yes, we want to catch it and log it
            log.error("Coroutine shutdown failed %s(%s): %r", log_awaitable(coro), log_args, error)
        took = time.time() - start
        timeout -= took

    return timeout


async def __shutdown_tasks(timeout: float) -> float:
    """Gather tasks and await for their finish with timeout"""
    log.debug("Shutting down tasks...")

    start = time.time()

    tasks = asyncio.all_tasks()
    tasks.discard(asyncio.current_task())  # don't touch ourselves
    for task in tasks:
        task.cancel()
    # We must await.sleep(0) here so task.cancel() is actually processed
    # and await in the task's coroutine is interrupted.
    await asyncio.sleep(0)

    for task in tasks.copy():
        if __task_failed(task, "%s await must be in try/except!") or task.done():
            # don't wait for failed task because they never gather
            tasks.remove(task)

    if not tasks:
        log.debug("No tasks to gracefully shutdown")
        took = time.time() - start
        return timeout - took

    try:
        log.debug(f"Waiting up to {timeout:.3f} s for tasks to finish...")
        await asyncio.wait_for(asyncio.gather(*tasks, return_exceptions=True), timeout=timeout)
        log.debug("All tasks finished within timeout")
    except asyncio.TimeoutError:
        log.warning("Nope, they get interrupted")

    for task in tasks:
        if not task.done():
            # Task didn't finish before timeout. Is it a bug in task code executed after cancel
            # or is the timeout too short to finish all needed work by task?
            log.error("Task did not finish gracefully: %s", log_awaitable(task))
            continue

        __task_failed(task, "%s didn't finish in shutdown timeout")

    took = time.time() - start
    return timeout - took


def __prepare_coroutines(*, after_tasks_cancel: bool) -> set[asyncio.Task]:
    """Prepare coroutines for shutdown"""
    tasks = set()
    for coro, args, when in __on_shutdown[::-1]:
        if when != after_tasks_cancel:
            continue

        try:
            coroutine = coro(*args)
        except Exception:  # noqa: BLE001 - yes, we want to catch it and log it
            log_args = ",".join(repr(arg) for arg in args)
            log.trace("Exception in %s(%s)", coro.__name__, log_args)
            continue

        if not asyncio.iscoroutine(coroutine):  # we may get just plain synchronous method
            continue

        try:
            tasks.add(create_task(coroutine, name="OnShutdown"))
        except Exception as error:  # noqa: BLE001 - yes, we want to catch it and log it
            log_args = ",".join(repr(arg) for arg in args)
            log.error("Unable to prepare for graceful shutdown %s(%s): %r", coro, log_args, error)

    return tasks


def __task_failed(task: asyncio.Task, canceled_msg: str) -> bool:
    """Log task result/exception and return if exception happened in its coroutine"""
    if not task.done():
        return False

    try:
        result = task.result()
    except asyncio.CancelledError:
        log.trace(canceled_msg, log_awaitable(task))
    except Exception as error:  # noqa: BLE001 - yes, we want to catch it and log it
        log.trace("%s in task %s: %s", error.__class__.__name__, log_awaitable(task), error)
    else:
        log.debug("Task %s finished with result: %r", log_awaitable(task), result)
        return False

    return True



"""INI like configuration loader"""

import configparser
from typing import Annotated, Optional, get_origin

from .annotations import ConfigOption, get_options, parse_list, set_options
from .exceptions import InvalidConfigImplementation


def loader(config_class: type[Annotated], filename: Optional[str] = None) -> Annotated:
    """INI like configuration loader"""
    config = get_options(config_class)

    if filename is None:
        return set_options(config_class(), config)

    conf = configparser.ConfigParser()
    conf.read(filename)
    for section_name, options in config.items():
        if not isinstance(options, dict):
            error = f'Class "{config_class.__name__}" can\'t have direct option "{section_name}"'
            raise InvalidConfigImplementation(error)

        for option_name, option in options.items():
            if not isinstance(option, ConfigOption):
                error = f'"{section_name}" can have only scalar attributes, not subsection "{option_name}"'
                raise InvalidConfigImplementation(error)

            if option.type is int:
                get = conf.getint
            elif option.type is float:
                get = conf.getfloat
            elif option.type is bool:
                get = conf.getboolean
            else:
                get = conf.get

            value = get(section_name, option_name, fallback=option.value)

            options[option_name] = parse_list(option.type, value) if get_origin(option.type) is list else value

    return set_options(config_class(), config)
